// name: AtomPositions.h
// author: J. Michael Word
// date written: 8/1/97
// purpose: Interface for AtomPositions

// **************************************************************
// NOTICE: This is free software and the source code is freely
// available. You are free to redistribute or modify under the
// conditions that (1) this notice is not removed or modified
// in any way and (2) any modified versions of the program are
// also available for free.
//               ** Absolutely no Warranty **
// Copyright (C) 1999 J. Michael Word
// **************************************************************

#ifndef ATOMPOSITIONS_H
#define ATOMPOSITIONS_H 1

#include "PDBrec.h"
#include "ResBlk.h"
#include "DotSph.h"
#include "BumperPoint.h"
#include "Mover.h"
#include "CliqueList.h"
#include "neighbors.h"
#include "utility.h"

class AtomPositions {
public:
   AtomPositions(int nblocks, bool onlyA, bool xplor, int nbCutoff,
		  float minRegHBcut, float minChargedHBcut,
		  float badBumpGapCut,
		  DotSphManager& dotBucket, float probeRadius,
		  float pmag, float occCutoff,
		  bool verbose, bool showOrientScore,
		  bool cliqueTicks, ostream& os)
      : _xyzBlocks(nblocks), _motionDesc(200),
        _onlyA(onlyA), _useXplorNames(xplor),
	_nBondCutoff(nbCutoff),
	_min_regular_hb_cutoff(minRegHBcut),
	_min_charged_hb_cutoff(minChargedHBcut),
	_bad_bump_gap_cutoff(badBumpGapCut),
	_dotBucket(dotBucket),
	_probeRadius(probeRadius), _pmag(pmag),
	_occupancyCuttoff(occCutoff),
	_outputNotice(verbose),
	_showOrientScore(showOrientScore),
	_cliqueTicks(cliqueTicks), _os(os) {}

  ~AtomPositions() {}

   int forceOrientations(const Stringclass& ofilename, List<Stringclass>& notes);

   void put(const PDBrec& r) {
      _xyzBlocks.put(LocBlk(r.loc()), r);
   }
   // move atom to new xyz pos.
   void reposition(const Point3d& prev, const PDBrec& r);

   Seq<PDBrec> neighbors(const Point3d& p,
                         Coord mindist, Coord maxdist) const {
      return ::neighbors(p, mindist, maxdist, _xyzBlocks);
   }

   void insertRot(const PDBrec& hr, const PDBrec& c1,
                  const PDBrec& c2, const PDBrec& c3,
		  bool doOHSH, bool doNH3, bool doMethyl);

   Seq<char> insertFlip(const ResBlk& rblk);
   void      insertFlip(const PDBrec& hr, Seq<char> alts);

   void doNotAdjust(const PDBrec& a);

   void finalizeMovers();

   CliqueList findCliques() const;

   int orientSingles(Seq<MoverPtr> singles);
   int orientClique(Seq<MoverPtr> clst, int limit); // returns -1 if abandoned
   int exhaustiveSearchOfClique(Seq<MoverPtr> clst);

   void describeChanges(NonConstListIter<PDBrec>& infoPtr, List<Stringclass>& notes);

   int numChanges() const { return _motionDesc.size(); }

   void manageMetals(const ResBlk& rblk);

   void generateWaterPhantomHs(Seq<PDBrec>& waters);

   double atomScore(const PDBrec& a, const Point3d& p,
		const Seq<PDBrec>& nearby, const Seq<PDBrec>& exclude,  // JSS: no need to copy
		const DotSph& dots, float prRadius, bool onlyBumps,
		float &bumpSubScore, float &hbSubScore, bool &hasBadBump);
private:
   AtomPositions(const AtomPositions& a);            // can't copy
   AtomPositions& operator=(const AtomPositions& a); // can't assign

   MultiDict<LocBlk, PDBrec> _xyzBlocks;
   Dict<Stringclass, Mover>       _motionDesc;
   Seq<BumperPoint>          _excludePoints;

   const bool                _onlyA;
   const bool                _useXplorNames;
   const int                 _nBondCutoff;
   const float               _min_regular_hb_cutoff;
   const float               _min_charged_hb_cutoff;
   const float               _bad_bump_gap_cutoff;
   DotSphManager&            _dotBucket;
   const float               _probeRadius;
   const float               _pmag;
   const float               _occupancyCuttoff;
   const bool                _outputNotice;
   const bool                _showOrientScore;
   const bool                _cliqueTicks;
   ostream&                  _os;
};
#endif
